package eu.jonahbauer.wizard.core.machine;

import eu.jonahbauer.wizard.common.messages.observer.ObserverMessage;
import eu.jonahbauer.wizard.common.messages.observer.UserInputMessage;
import eu.jonahbauer.wizard.common.messages.player.PickTrumpMessage;
import eu.jonahbauer.wizard.common.messages.player.PlayCardMessage;
import eu.jonahbauer.wizard.common.messages.player.PlayerMessage;
import eu.jonahbauer.wizard.common.messages.player.PredictMessage;
import eu.jonahbauer.wizard.common.model.card.Card;
import eu.jonahbauer.wizard.core.messages.Observer;
import lombok.Getter;
import lombok.RequiredArgsConstructor;
import lombok.Setter;
import org.junit.jupiter.api.Assertions;

import java.util.*;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import static org.junit.jupiter.api.Assertions.*;

/**
 * A queue of {@link PlayerMessage}s that will automatically be sent to a {@link Game} when
 * {@linkplain UserInputMessage input is required}. The {@code PlayerMessage}s are ordered and it is asserted that
 * the incoming {@code UserInputMessage}s match the given {@code PlayerMessage}s.
 */
public class MessageQueue implements Observer {
    private final ExecutorService executor = Executors.newSingleThreadExecutor();
    private final LinkedList<QueuedMessage> messages = new LinkedList<>();
    @Setter
    private Game game;

    public MessageQueue add(UUID player, UserInputMessage.Action action, PlayerMessage message) {
        messages.add(new QueuedMessage(player, action, message));
        return this;
    }

    public MessageQueue addCard(UUID player, Card card) {
        return add(player, UserInputMessage.Action.PLAY_CARD, new PlayCardMessage(card));
    }

    public MessageQueue addPrediction(UUID player, int prediction) {
        return add(player, UserInputMessage.Action.MAKE_PREDICTION, new PredictMessage(prediction));
    }

    public MessageQueue addChangePrediction(UUID player, int prediction) {
        return add(player, UserInputMessage.Action.CHANGE_PREDICTION, new PredictMessage(prediction));
    }

    public MessageQueue addPickTrump(UUID player, Card.Suit trumpSuit) {
        return add(player, UserInputMessage.Action.PICK_TRUMP, new PickTrumpMessage(trumpSuit));
    }

    public MessageQueue addCards(List<UUID> players, Map<UUID, List<Card>> cards, int startPlayer) {
        Map<UUID, Integer> nextCards = new HashMap<>();
        int size = players.size();
        boolean changed = true;
        while (changed) {
            changed = false;
            for (int i = 0; i < size; i++) {
                var player = players.get((i + startPlayer) % size);
                var hand = cards.get(player);

                int next = nextCards.getOrDefault(player, 0);

                if (next < hand.size()) {
                    nextCards.merge(player, 1, Integer::sum);
                    var card = hand.get(next);
                    addCard(player, card);
                    changed = true;
                }
            }
        }
        return this;
    }

    public MessageQueue assertThrows(Class<? extends Exception> exception) {
        messages.getLast().setException(exception);
        return this;
    }

    public void doNotify(ObserverMessage om) {
        try {
            System.out.println(om);
            if (om instanceof UserInputMessage message) {
                UUID player = message.getPlayer();
                while (true) {
                    assertFalse(messages.isEmpty(), "User input is required but none is provided.");

                    var queued = messages.poll();

                    var queuedPlayer = queued.getPlayer();
                    var queuedAction = queued.getAction();
                    var queuedMessage = queued.getMessage();
                    var exception = queued.getException();

                    if (exception == null) {
                        assertEquals(queuedPlayer, player);
                        assertEquals(queuedAction, message.getAction());
                    }

                    System.out.println(queuedPlayer + ": " + queuedMessage);
                    if (exception != null) {
                        var executionException = Assertions.assertThrows(
                                ExecutionException.class,
                                () -> game.onMessage(queuedPlayer, queuedMessage).get(),
                                "Excepted exception for message " + queuedMessage + " from player " + queuedPlayer + "."
                        );
                        assertInstanceOf(
                                exception,
                                executionException.getCause(),
                                "Excepted exception for message " + queuedMessage + " from player " + queuedPlayer + "."
                        );
                    } else {
                        Assertions.assertDoesNotThrow(() -> game.onMessage(queuedPlayer, queuedMessage).get());
                        return;
                    }
                }
            }
        } catch (Throwable t) {
            game.finish(t);
        }
    }

    @Override
    public void notify(ObserverMessage message) {
        executor.execute(() -> doNotify(message));
    }

    @Override
    public void notify(UUID player, ObserverMessage message) {

    }

    @Getter
    @Setter
    @RequiredArgsConstructor
    private static class QueuedMessage {
        private final UUID player;
        private final UserInputMessage.Action action;
        private final PlayerMessage message;
        private Class<? extends Exception> exception;
    }
}
